/*
hinit.c - Release 0.2.0  Hash initialize database
Copyright (C) 2003-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define RCDHDR        4
#define MAXRCDS 9999999
#define MAXRCDLEN 65536

int HClsInit(hndl)
int hndl;
   {
   int rslt;
   rslt = close(hndl);
   if (rslt < 0)
      {
      perror("HClsInit: close error");
      return(SYS_ERR + errno);
      } /* if close error */
   return(0);
   } /* HClsInit */

int HGetBlkInit(hndl,rcdlen,rcd)
int hndl;
int rcdlen;
unsigned char *rcd;
   {
   int rdlen;
   rdlen = read(hndl,rcd,rcdlen);
   if (rdlen == rcdlen)
      return(0);
   else if (rdlen < 0)
      {
      fprintf(stderr,"HGetBlkInit: errno %d\n", errno);
      perror("Read error");
      HClsInit(hndl);
      return(SYS_ERR + errno);
      } /* read error */
   else
      {
      fprintf(stderr,"HGetBlkInit: read error "
	 "(rdlen %d)\n", rdlen);
      HClsInit(hndl);
      return(HASH_ERR + 107);
      } /* read error */
   } /* HGetBlkInit */

int HPutBlkInit(hndl,rcdlen,rcd)
int hndl;
int rcdlen;
unsigned char *rcd;
   {
   int wrtlen;
   wrtlen = write(hndl,rcd,rcdlen);
   if (wrtlen == rcdlen)
      return(0);
   else if (wrtlen < 0)
      {
      fprintf(stderr,"HPutBlkInit: errno %d\n", errno);
      perror("Write error");
      HClsInit(hndl);
      return(SYS_ERR + errno);
      } /* write error */
   else
      {
      fprintf(stderr,"HPutBlkInit: write error "
	 "(wrtlen %d)", wrtlen);
      HClsInit(hndl);
      return(HASH_ERR + 108);
      } /* write error */
   return(0);
   } /* HPutBlkInit */

int HBldHdrInit(hndl,totrcds,
   rcdlen,keylen,keyofst)
int hndl;
int totrcds;
int rcdlen;
int keylen;
int keyofst;
   {
   int i;
   int rslt;
   int hdrsz;
   int nodelen;
   h_nodefmt *node;
   hashfmt *blk;
   unsigned char *p,*q;
   rslt = 0;
   hdrsz = HDRLEN;
   blk = (hashfmt *) malloc(hdrsz+16);
   if (blk == NULL)
      {
      fprintf(stderr,"HBldHdrInit: malloc error\n");
      HClsInit(hndl);
      return(HASH_ERR + 110);
      } /* malloc err */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + hdrsz;
   while (p < q) *p++ = '\0';
   strcpy(blk->signature,H_SIGNATURE);
   blk->totrcds = totrcds;
   blk->rcdlen  = rcdlen;
   blk->keylen  = keylen;
   blk->keyofst = keyofst;
   blk->rcdhdr  = RCDHDR;
   blk->hdrlen  = HDRLEN;
   nodelen = rcdlen + RCDHDR;
   rslt = HPutBlkInit(hndl,HDRLEN,blk);
   if (rslt) return(rslt);
   /* blank records for hashing area */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + nodelen;
   while (p < q) *p++ = '\0';
   node = (h_nodefmt *) blk;
   node->next = UINT_MAX;
   i = totrcds;
   while (i--)
      {
      rslt = HPutBlkInit(hndl,nodelen,node);
      if (rslt) break;
      } /* initialize database */
   free(blk);
   return(rslt);
   } /* HBldHdrInit */

int HOpnInitOld(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   unsigned char buf[32];
   *hndl = -1;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl < 0)
      {
      fprintf(stderr,"HOpnInitOld: %s: "
	 "open error\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      } /* open err */
   rslt = lockf(oldhndl,F_TEST,0);
   if (rslt < 0)
      {
      fprintf(stderr,"HOpnInitOld: lock test "
         "error on %s\n", dbname);
      perror("Lock test error");
      HClsInit(oldhndl);
      return(SYS_ERR + errno);
      } /* if lock test failed */
   rslt = HGetBlkInit(oldhndl,16,buf);
   if (rslt) return(rslt);
   rslt = strncmp((char *) buf,H_SIGNATURE,16); 
   if (rslt)
      {
      fprintf(stderr,"HOpnInitOld: filename %s "
         "for non-hash file\n",
         dbname);
      HClsInit(oldhndl);
      return(HASH_ERR + 106);
      } /* not hash file */
   else
      {
      rslt = HClsInit(oldhndl);
      if (rslt) return(rslt);
      } /* close after good signature */
   *hndl = open(dbname,O_RDWR|O_TRUNC);
   if (*hndl < 0)
      {
      fprintf(stderr,"HOpnInitOld: %s: "
	 "open read/write\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      } /* if open error */
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"HOpnInitOld: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      HClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* HOpnInitOld */

int HOpnInitNew(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl > 0)
      {
      fprintf(stderr,"HOpnInitNew: %s: "
	 "file exists\n",
         dbname);
      HClsInit(oldhndl);
      return(HASH_ERR + 109);
      } /* open err */
   *hndl = open(dbname,
      O_RDWR|O_CREAT|O_EXCL,
      S_IRUSR | S_IWUSR);
   if (*hndl < 0)
      {
      fprintf(stderr,"HOpnInitNew: %s: "
	 "open read/write\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      }
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"HOpnInitNew: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      HClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* HOpnInitNew */

int HOpnInitAll(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   unsigned char buf[32];
   *hndl = -1;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl < 0)
      {
      if (errno == ENOENT);
      else
	 {
         fprintf(stderr,"HOpnInitAll: %s: "
	    "open error\n",
            dbname);
         perror("Open error");
         return(SYS_ERR + errno);
	 } /* serious open err */
      } /* open err */
   else if (oldhndl > 0)
      {
      rslt = lockf(oldhndl,F_TEST,0);
      if (rslt < 0)
         {
         fprintf(stderr,"HOpnInitAll: lock test "
            "error on %s\n", dbname);
         perror("Lock test error");
	 HClsInit(oldhndl);
         return(SYS_ERR + errno);
         } /* if lock test failed */
      rslt = HGetBlkInit(oldhndl,16,buf);
      if (rslt) return(rslt);
      rslt = strncmp((char *) buf,H_SIGNATURE,16); 
      if (rslt)
	 {
	 fprintf(stderr,"HOpnInitAll: filename %s "
	    "for non-hash file\n",
	    dbname);
	 HClsInit(oldhndl);
	 return(HASH_ERR + 115);
	 } /* not hash file */
      else
	 {
	 rslt = HClsInit(oldhndl);
	 if (rslt) return(rslt);
	 } /* close after good signature */
      } /* if good open rdonly */
   *hndl = open(dbname,O_RDWR|O_TRUNC);
   if (*hndl < 0)
      {
      if (errno == ENOENT)
	 {
	 fprintf(stderr,"Creating %s\n", dbname);
         *hndl = open(dbname,
	    O_RDWR|O_CREAT|O_EXCL,
            S_IRUSR | S_IWUSR);
         if (*hndl < 0)
            {
	    fprintf(stderr,"HOpnInitAll: %s: "
	       "open read/write\n",
	       dbname);
	    perror("Open error");
            return(SYS_ERR + errno);
            }
	 } /* new database */
      else
	 {
	 fprintf(stderr,"HOpnInitAll: %s: "
	    "open read/write\n",
	    dbname);
	 perror("Open error");
         return(SYS_ERR + errno);
	 } /* genuine open error */
      } /* if open error */
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"HOpnInitAll: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      HClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* HOpnInitAll */

int HOpnInit(dbname,truncflag,
   totrcds,rcdlen,keylen,keyofst)
char *dbname;
int truncflag;
int totrcds;
int rcdlen;
int keylen;
int keyofst;
   {
   int rslt;
   int hndl;
   if (truncflag == INITOLD)
      rslt = HOpnInitOld(dbname,&hndl);
   else if (truncflag == INITNEW)
      rslt = HOpnInitNew(dbname,&hndl);
   else if (truncflag == INITALL)
      rslt = HOpnInitAll(dbname,&hndl);
   if (rslt) return(rslt);

   rslt = HBldHdrInit(hndl,totrcds,
      rcdlen,keylen,keyofst);
   if (rslt) return(rslt);
   return(HClsInit(hndl));
   } /* HOpnInit */

int HValidateInit(dbname,truncflag,
   totrcds,rcdlen,keylen,keyofst)
char *dbname;
int truncflag;
int totrcds;
int rcdlen;
int keylen;
int keyofst;
   {
   if (keylen > rcdlen)
      {
      fprintf(stderr,"hinit: key-length %d "
         "exceeds record-length %d\n",
	 keylen, rcdlen);
      return(HASH_ERR + 112);
      }
   if (keyofst > rcdlen - keylen)
      {
      fprintf(stderr,"hinit: key-offset %d "
         "exceeds the record-length %d\n",
	 keyofst, rcdlen);
      fprintf(stderr,"    minus the key-length %d\n",
	 keylen);
      return(HASH_ERR + 113);
      }
   return(HOpnInit(dbname,truncflag,
      totrcds,rcdlen,keylen,keyofst));
   } /* HValidateInit */

int hinit(dbname,truncflag,
   totrcds,rcdlen,keylen,keyofst)
char *dbname;
int truncflag;
int totrcds;
int rcdlen;
int keylen;
int keyofst;
   {
   int namelen;
   namelen =  strlen(dbname);
   if (namelen < 1 || namelen > 128)
      {
      fprintf(stderr,"hinit: wrong "
	 "length dbname %s\n",
	 dbname);
      return(HASH_ERR + 116);
      } /* dbname wrong length */
   if (truncflag == INITNEW
      || truncflag == INITOLD
      || truncflag == INITALL);
   else
      {
      fprintf(stderr,"hinit: invalid "
	 "truncate flag %d\n",
	 truncflag);
      return(HASH_ERR + 114);
      } /* bad truncflag */
   if (totrcds < 1 || totrcds > MAXRCDS)
      {
      fprintf(stderr,"hinit: invalid "
         "totrcds %d\n",
         totrcds);
      return(HASH_ERR + 101);
      }
   if (totrcds != hprime(totrcds))
      {
      fprintf(stderr,"hinit: totrcds "
	 "is not prime\n");
      return(HASH_ERR + 117);
      }
   if (rcdlen < 1 || rcdlen > MAXRCDLEN)
      {
      fprintf(stderr,"hinit: invalid "
         "record-length %d\n",
         rcdlen);
      return(HASH_ERR + 103);
      }
   if (keylen < 1 || keylen > MAXRCDLEN)
      {
      fprintf(stderr,"hinit: invalid "
         "key-length %d\n",
         keylen);
      return(HASH_ERR + 104);
      }
   if (keyofst < 0 || keyofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"hinit: invalid "
         "key-offset %d\n",
         keyofst);
      return(HASH_ERR + 105);
      }
   return(HValidateInit(dbname,truncflag,
      totrcds,rcdlen,keylen,keyofst));
   } /* hinit */
